import matplotlib.pyplot as plt
from .. import __user_conf_dir
from ..parser import parseDatasetConfig
from ..parser import save 
import configparser
from ..parser.combine import runGroupData
from .fitter import plotFit 
from .transformer import transforms
from .decorator import decorations
import importlib
import os

# Search for default plot types
roots = []
modFilenames = []
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/plot_types"))
roots.append(root)
modFilenames.append(filenames)
# Search for custom plot types
if os.path.isdir(os.path.expanduser(__user_conf_dir+"/plot_types")):
    root, _, filenames = next(os.walk(os.path.expanduser(__user_conf_dir+"/plot_types")))
    roots.append(root)
    modFilenames.append(filenames)
# Import all plot types
plotTypes = {}
for i in range(len(roots)):
    for filename in modFilenames[i]:
        plotTypeName = filename.split(".")[0]
        if plotTypeName[0:2] == "__":
            # Skip __init__.py and similar commands
            continue
        spec = importlib.util.spec_from_file_location("compphysutils.graphics.plot_types."+plotTypeName, roots[i]+"/"+filename)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        plotTypes[plotTypeName] = mod.plot

class CyclicIterator:
    def __init__(self, cycle=[]):
        self.singleCycle = cycle
        self.cycleLen = len(cycle)
        self.currentIndex = 0

    def __iter__(self):
        return self

    def __next__(self):
        returnVal = self.singleCycle[self.currentIndex % self.cycleLen]
        self.currentIndex += 1
        return returnVal

class ColorIterator(CyclicIterator):
    def __init__(self, singleCycle="b"):
        # Change the format if necessary
        listOfColors = singleCycle.split()
        for i in range(len(listOfColors)):
            if listOfColors[i].find(",") >= 0:
                listOfColors[i] = tuple(map(float, listOfColors[i].split(",")))
        super().__init__(listOfColors)

class LinestyleIterator(CyclicIterator):
    def __init__(self, singleCycle="-"):
        super().__init__(singleCycle.split())

class MarkerstyleIterator(CyclicIterator):
    def __init__(self, singleCycle="o"):
        super().__init__(singleCycle.split())

def plot(datasets, plotType="scatter", axes=False, figure=False, **plotOptions):
    if not figure:
        figure = plt.gcf()
    if not axes:
        axes = figure.gca()
    if plotType in plotTypes:
        plotTypes[plotType](datasets, axes, figure=figure, **plotOptions)
    else:
        # Defaults to scatter
        plotTypes["scatter"](datasets, axes, **plotOptions)
    axes.set_xlabel(plotOptions["xlabel"])
    axes.set_ylabel(plotOptions["ylabel"])
    if plotOptions["xlim"]:
        axes.set_xlim(*plotOptions["xlim"])
    if plotOptions["ylim"]:
        axes.set_ylim(*plotOptions["ylim"])
    # Insert tick labels
    if plotOptions["xticks"]:
        #axes.set_xticks(plotOptions["xticks"][0], labels=plotOptions["xticks"][1])
        axes.set_xticks(plotOptions["xticks"][0])
        axes.set_xticklabels(plotOptions["xticks"][1])
    if plotOptions["yticks"]:
        axes.set_yticks(plotOptions["yticks"][0])
        axes.set_yticklabels(plotOptions["yticks"][1])
    return axes

def fromConfig(configFileName, axes=False, figure=False, datasets={}):
    axesGiven = False
    if not figure:
        figure = plt.gcf()
    if axes:
        axesGiven = True
    cfg = configparser.ConfigParser()
    cfg.read(configFileName)
    datasets.update(runGroupData(cfg, datasets, configFileName))
    # Now, run any transform commands
    if "transform" in cfg["plot"]:
        transformCommands = cfg["plot"].get("transform").split("\n")
        for commandLine in transformCommands:
            commandSplitLine = commandLine.split()
            commandName = commandSplitLine[0]
            datasets = transforms[commandName](datasets, commandSplitLine[1:])
    if "savepoint" in cfg["plot"]:
        save(cfg["plot"].get("savepoint"), "transform", datasets)
    # Now, datasets are complete, and we can read the plot group
    # Also include options that are set directly via type - should be reserved for options that are not usable for many plot types
    graphTypeSplit = cfg["plot"].get("type", "scatter").split()
    graphType = graphTypeSplit[0]
    colCoords = cfg["plot"].get("cols", False)
    if colCoords:
        colCoords = colCoords.split("\n")
    else:
        colCoords = []
    for i in range(len(colCoords)):
        colCoords[i] = colCoords[i].split()
    chosenDatasets = []
    for i in range(len(colCoords)):
        chosenDatasets.append([])
        for j in range(0,len(colCoords[i]),2):
            try:
                chosenDatasets[i].append(datasets[colCoords[i][j]][int(colCoords[i][j+1])])
            except IndexError:
                raise IndexError("Cannot create plot for coordinates "+colCoords[i][j]+" "+colCoords[i][j+1])
    plotOptions = {}
    plotOptions["plotArgString"] = graphTypeSplit[1:]
    plotOptions["legend"] = cfg["plot"].getboolean("legend", True)
    plotOptions["legend-pos"] = cfg["plot"].get("legend-pos", "best").split("\n")
    plotOptions["legend-cols"] = list(map(int, cfg["plot"].get("legend-cols", "1").split("\n")))
    if "xlim" in cfg["plot"]:
        plotOptions["xlim"] = list(map(float, cfg["plot"].get("xlim").split()))
    else:
        plotOptions["xlim"] = False
    if "ylim" in cfg["plot"]:
        plotOptions["ylim"] = list(map(float, cfg["plot"].get("ylim").split()))
    else:
        plotOptions["ylim"] = False
    if "font" in cfg["plot"]:
        plt.rcParams["font.family"] = cfg["plot"].get("font", "sans")
    if "font-size" in cfg["plot"]:
        plt.rcParams["font.size"] = int(cfg["plot"].get("font-size", 12))
    if "mathfont" in cfg["plot"]:
        plt.rcParams["mathtext.fontset"] = cfg["plot"].get("mathfont", "cm")
    figure.set_size_inches(
            float(cfg["plot"].get("fig-width-inches", 6.4)),
            float(cfg["plot"].get("fig-height-inches", 4.8))
    )
    plotOptions["colorCycle"] = cfg["plot"].get("colorCycle", "b")
    plotOptions["colorCycle"] = ColorIterator(plotOptions["colorCycle"])
    plotOptions["linestyleCycle"] = cfg["plot"].get("linestyleCycle", "solid")
    plotOptions["linestyleCycle"] = LinestyleIterator(plotOptions["linestyleCycle"])
    plotOptions["markerstyleCycle"] = cfg["plot"].get("markerstyleCycle", "o")
    plotOptions["markerstyleCycle"] = LinestyleIterator(plotOptions["markerstyleCycle"])
    plotOptions["xlabel"] = cfg["plot"].get("xlabel", None)
    plotOptions["ylabel"] = cfg["plot"].get("ylabel", None)
    plotOptions["figfile"] = cfg["plot"].get("figfile", False)
    # Dataset labels
    plotOptions["datasetLabels"] = cfg["plot"].get("labels", False)
    if plotOptions["datasetLabels"]:
        plotOptions["datasetLabels"] = plotOptions["datasetLabels"].split("\n")
        if len(plotOptions["datasetLabels"]) < len(colCoords):
            toAdd = len(colCoords) - len(plotOptions["datasetLabels"])
            for i in range(toAdd):
                plotOptions["datasetLabels"].append(None)
    # Arguments supplied are the dataset name, convert it to a dataset that is then plotted
    for ticksName in ["xticks", "yticks"]:
        plotOptions[ticksName] = cfg["plot"].get(ticksName, False)
        if plotOptions[ticksName] and (not cfg["plot"].get("hide-"+ticksName, False)):
            plotOptions[ticksName] = datasets[plotOptions[ticksName]]
        elif cfg["plot"].get("hide-"+ticksName, False):
            plotOptions[ticksName] = [[],[]]
    axes = plot(chosenDatasets, graphType, axes=axes, figure=figure, **plotOptions)
    # If the axes are hidden, hide them
    if cfg["plot"].get("hide-axes", False):
        axes.set_axis_off()
    # If fit is present, handle it
    fitIndex = 0
    if cfg["plot"].get("fit", False):
        fitLabels = cfg["plot"].get("fit-labels", False)
        numFits = len(cfg["plot"].get("fit").split("\n"))
        if not fitLabels:
            fitLabels = [False] * numFits
        else:
            fitLabels = fitLabels.split("\n")
        if len(fitLabels) < numFits:
            fitLabels += [False] * (numFits - len(fitLabels))
        prevFitParams = []
        fitColorIterator = ColorIterator(cfg["plot"].get("fit-colorCycle", "tab:blue tab:orange tab:green tab:cyan"))
        for allFitArgs in cfg["plot"].get("fit").split("\n"):
            fitArgs = allFitArgs.split()
            # TODO : Fit args?
            prevFitParams += list(plotFit(
                chosenDatasets[int(fitArgs[1])],
                fitArgs[0],
                axes,
                fitPoints=int(cfg["plot"].get("fit-points", 100)),
                fitLabel=fitLabels[fitIndex],
                showParams=cfg["plot"].getboolean("fit-show-params", True),
                showError=cfg["plot"].getboolean("fit-show-error", True),
                fitColorCycle=fitColorIterator,
                paramsPlacement=cfg["plot"].get("params-placement", False),
                paramsOffset=len(prevFitParams),
                xMin=float(cfg["plot"].get("fit-xmin", False)),
                xMax=float(cfg["plot"].get("fit-xmax", False)),
                dirtyRun=cfg["plot"].getboolean("fit-dirty-run", False)
                ))
            fitIndex += 1
    # Handle decorations for main axes
    if cfg["plot"].get("decorate", False):
        decorationCommands = cfg["plot"].get("decorate").split("\n")
        for decorationArgs in decorationCommands:
            decorationSplit = decorationArgs.split()
            axes, datasets = decorations[decorationSplit[0]](axes, datasets, decorationSplit[1:])
    # Move the legend render here, even after decorations (which can also be annotated)
    if plotOptions["legend"]:
        if len(plotOptions["legend-pos"]) == 1:
            # Apply the same location and number of columns for each legend
            legendArgs = plotOptions["legend-pos"][0].split()
            if len(legendArgs) == 4:
                axes.legend(loc=" ".join(legendArgs[:-2]), ncol=plotOptions["legend-cols"][0], bbox_to_anchor=tuple(map(float, legendArgs[-2:])))
            else:
                axes.legend(loc=plotOptions["legend-pos"][0], ncol=plotOptions["legend-cols"][0])
        else:
            # Split the legend - now need that the number of lines and legend entries are the same
            handles, labels = axes.get_legend_handles_labels()
            # First, organize line handles into list with the same positions - use dictionary
            legendsOrganized = {}
            currentLocation = 0
            for i in range(len(handles)):
                if plotOptions["legend-pos"][i] in legendsOrganized:
                    legendsOrganized[plotOptions["legend-pos"][i]]["handles"].append(handles[i])
                else:
                    # Create new descriptor object
                    legendsOrganized[plotOptions["legend-pos"][i]] = {}
                    legendsOrganized[plotOptions["legend-pos"][i]]["handles"] = [handles[i]]
                    if currentLocation < len(plotOptions["legend-cols"]):
                        legendsOrganized[plotOptions["legend-pos"][i]]["cols"] = plotOptions["legend-cols"][currentLocation]
                        currentLocation += 1
                    else:
                        legendsOrganized[plotOptions["legend-pos"][i]]["cols"] = 1
            legendArtists = []
            for legendLoc in legendsOrganized:
                if len(legendLoc.split()) <= 2:
                    # Only loc is provided
                    legendArtists.append(axes.legend(handles=legendsOrganized[legendLoc]["handles"], loc=legendLoc, ncol=legendsOrganized[legendLoc]["cols"]))
                else:
                    # Loc and bbox are provided
                    legendLocOnly = " ".join(legendLoc.split()[:-2])
                    legendBBOXOnly = tuple(map(float, legendLoc.split()[-2:]))
                    legendArtists.append(axes.legend(handles=legendsOrganized[legendLoc]["handles"], loc=legendLocOnly, bbox_to_anchor=legendBBOXOnly, ncol=legendsOrganized[legendLoc]["cols"]))
            # Finally, add overwritten artists back to the axes
            for i in range(len(legendArtists)-1):
                axes.add_artist(legendArtists[i])
    # If an inset directive is present, add an inset to the current axes
    if cfg["plot"].get("inset", False):
        insetLines = cfg["plot"].get("inset").split("\n")
        for i in range(len(insetLines)):
            insetArgs = insetLines[i].split()
            # Arguments are xpos, ypos, xwidth, ywidth
            insetAxes = axes.inset_axes(list(map(float, insetArgs[1:])))
            fromConfig(insetArgs[0], axes=insetAxes, figure=figure, datasets=datasets)
    if cfg["plot"].get("overlay", False):
        # Split by any whitespace
        overlayLines = cfg["plot"].get("overlay").split()
        for i in range(len(overlayLines)):
            # Apply a second graph on top of this one
            fromConfig(overlayLines[i], axes=axes, figure=figure, datasets=datasets)
    if cfg["plot"].get("twinx", False):
        # Plot another dataset sharing the same x axis but different y axis
        # Always, only a single twinx makes sense - provide no arguments
        fromConfig(cfg["plot"].get("twinx"), axes=axes.twinx(), figure=figure, datasets=datasets)
    # If axes are provided, assume figure is printed somewhere else
    # TODO : Is this a reasonable assumption?
    if axesGiven:
        return True
    if plotOptions["figfile"]:
        # Can output the same figure in different formats
        for figFileName in plotOptions["figfile"].split():
            plt.savefig(figFileName, bbox_inches="tight")
    else:
        plt.show()
    return figure



import argparse

joinAP = argparse.ArgumentParser()
joinAP.add_argument("new_name", help="Name of the joined dataset.")
joinAP.add_argument("datasets_to_join", nargs="+", help="Datasets which will constitute the columns of the new dataset.")
def command(datasets, commandArgs):
    args = joinAP.parse_args(commandArgs)
    newDataset = []
    for elementName in args.datasets_to_join:
        for i in range(len(datasets[elementName])):
            newDataset.append(datasets[elementName][i])
    datasets[args.new_name] = newDataset
    return datasets



import argparse
import matplotlib.image
import numpy as np

def initParserObjects(parserArgs):
    ap = argparse.ArgumentParser()
    ap.add_argument("-v", action="store_true", help="Whether to output debug information. Default : False")
    args = ap.parse_args(parserArgs)
    return [args]

def file(filename, args):
    image = np.array(matplotlib.image.imread(filename))
    # Output dimension data in verbose
    if args.v:
        print("Image box size : ({:d},{:d})".format(image.shape[0], image.shape[1]))
    # Flatten image channels
    flatImage = []
    for i in range(image.shape[2]):
        flatImage.append(image[:,:,i].flatten())
    return flatImage



import argparse

gapReadAP = argparse.ArgumentParser()
gapReadAP.add_argument("--fractional", default=False, action="store_true", help="Instead of integer occupation, assume fractional occupation is present, and set LUMO to first level, where occupation is less than one electron.")
def command(datagroups, argList):
    # Prepared for reading of gap - expects datagrous [energy, occupation]
    LUMO = 0
    HOMO = 0
    args = gapReadAP.parse_args(argList)
    for i in range(len(datagroups[0])):
        # Data are ordered by energy
        if not args.fractional:
            if datagroups[1][i] == 0:
                LUMO = datagroups[0][i]
                HOMO = datagroups[0][i-1]
                break
        else:
            if datagroups[1][i] < 1.0:
                LUMO = datagroups[0][i]
                HOMO = datagroups[0][i-1]
                break
    return [[LUMO - HOMO]]



import argparse
import math

logTransAP = argparse.ArgumentParser()
logTransAP.add_argument("dataset_name", help="Name of the dataset.")
logTransAP.add_argument("column_index", type=int, help="Index of the column in the dataset.")
logTransAP.add_argument("--err", type=int, help="Do also an error calculation, on a different column.", default=False)
logTransAP.add_argument("--base", type=float, default=math.e, help="Base of the logarithm.")
def command(datasets, argsString):
    args = logTransAP.parse_args(argsString)
    for i in range(len(datasets[args.dataset_name][args.column_index])):
        origX = datasets[args.dataset_name][args.column_index][i]
        datasets[args.dataset_name][args.column_index][i] = math.log(origX, args.base)
        if args.err:
            datasets[args.dataset_name][args.err][i] = abs(math.log(origX + datasets[args.dataset_name][args.err][i], args.base) - math.log(origX, args.base))
    return datasets



import argparse

lineAP = argparse.ArgumentParser()
lineAP.add_argument("--vert", action="store_true", default=False, help="Draw vertical line instead of horizontal line.")
# Instead of parsing the kwargs straight, do argument conversion - is more backend agnostic
lineAP.add_argument("--color", default="k", help="Color of the line. [default : k = black]")
lineAP.add_argument("--style", default="-", help="Style of the line. [default : - = solid]")
lineAP.add_argument("coord", type=float, default=0.0, help="Coordinate of the line to be drawn [default : data coordinates]")
# TODO : Transforms to other coordinate types

def command(axes, datasets, argString):
    args = lineAP.parse_args(argString)
    # Draw a line
    if args.vert:
        axes.axvline(args.coord, ls=args.style, c=args.color)
    else:
        axes.axhline(args.coord, ls=args.style, c=args.color)
    return axes, datasets



import re
import argparse

class ColsReference:
    # Object that ensures that only the converged eigenvalues are read
    def __init__(self, parserArgs=["0"]):
        self.argsReader = argparse.ArgumentParser(description="Parses the arguments for the parser itself.")
        self.argsReader.add_argument("--comment", help="Character to be used as indicator of comment. [default : #]", default="#")
        self.argsReader.add_argument("cols", nargs="+", type=int, help="Indices of columns to read.")
        self.args = self.argsReader.parse_args(parserArgs)

    def continueReading(self, line):
        if line[0] == self.args.comment:
            return False
        # Not commented, read
        # TODO : Different var types
        splitLine = line.split()
        retVal = []
        for i in self.args.cols:
            retVal.append(float(splitLine[i]))
        return retVal

def line(line, CR):
    # Reading
    result = CR.continueReading(line)
    if not result:
        # Stop reading
        return False
    # Real result present, convert to floats and return
    return result

def initParserObjects(parserArgs):
    return [ColsReference(parserArgs)]

argDefaults = "0"

def writeLine(datarow):
    datalist = list(map(str, datarow))
    return " ".join(datalist)



from .post_processor import postProcessCommands
from .savepoint import handleSavepoints
from .. import __user_conf_dir
import configparser
import os
import sys
import importlib
import re

parserModules = {}
# Check for custom parser modules
customParsers = []
if os.path.isdir(os.path.expanduser(__user_conf_dir+"/parsers")):
    root, _, filenames = next(os.walk(os.path.expanduser(__user_conf_dir+"/parsers")))
    customParsers = list(map(lambda f: root+"/"+f, filenames))
# Check for default parser modules
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/parsers"))
defaultParsers = list(map(lambda f: root+"/"+f, filenames))
# Ideally, custom parsers should overwrite default parsers, not sure if this works
rexp = re.compile(".*/(.*)\.py")
for parserFileName in (defaultParsers + customParsers):
    parserType = rexp.search(parserFileName).group(1)
    if parserType[0:2] == "__":
        # Not a parser, but just a init or other module
        continue
    spec = importlib.util.spec_from_file_location("compphysutils.parser.parsers."+parserType, parserFileName)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    parserModules[parserType] = mod
readLineFunctions = {}
writeLineFunctions = {}
readFileFunctions = {}
writeFileFunctions = {}
readHeaderFunctions = {}
writeHeaderFunctions = {}
readFooterFunctions = {}
writeFooterFunctions = {}
readArgDefaults = {}
writeArgDefaults = {}
initReadObjects = {}
initWriteObjects = {}
for parserName in parserModules:
    # Prefer line-by-line reading when possible
    if hasattr(parserModules[parserName], "line"):
        readLineFunctions[parserName] = parserModules[parserName].line
    elif hasattr(parserModules[parserName], "file"):
        readFileFunctions[parserName] = parserModules[parserName].file
        readLineFunctions[parserName] = False
    else:
        # Neither line nor file reading utility provided, exception
        raise ModuleNotFoundError("Neither file nor line functions found for parser "+parserName+"!")
    if hasattr(parserModules[parserName], "writeLine"):
        writeLineFunctions[parserName] = parserModules[parserName].writeLine
    else:
        writeLineFunctions[parserName] = False
    if hasattr(parserModules[parserName], "writeFile"):
        writeFileFunctions[parserName] = parserModules[parserName].writeFile
    else:
        writeFileFunctions[parserName] = False
    if hasattr(parserModules[parserName], "argDefaults"):
        readArgDefaults[parserName] = parserModules[parserName].argDefaults
    else:
        readArgDefaults[parserName] = ""
    if hasattr(parserModules[parserName], "writeArgDefaults"):
        writeArgDefaults[parserName] = parserModules[parserName].writeArgDefaults
    else:
        writeArgDefaults[parserName] = ""
    if hasattr(parserModules[parserName], "initParserObjects"):
        initReadObjects[parserName] = parserModules[parserName].initParserObjects
    else:
        initReadObjects[parserName] = False
    if hasattr(parserModules[parserName], "initWriterObjects"):
        initWriteObjects[parserName] = parserModules[parserName].initWriterObjects
    else:
        initWriteObjects[parserName] = False
    if hasattr(parserModules[parserName], "readHeaders"):
        readHeaderFunctions[parserName] = parserModules[parserName].readHeaders
    else:
        readHeaderFunctions[parserName] = False
    if hasattr(parserModules[parserName], "readFooters"):
        readFooterFunctions[parserName] = parserModules[parserName].readFooters
    else:
        readFooterFunctions[parserName] = False
    if hasattr(parserModules[parserName], "writeHeaders"):
        writeHeaderFunctions[parserName] = parserModules[parserName].writeHeaders
    else:
        writeHeaderFunctions[parserName] = False
    if hasattr(parserModules[parserName], "writeFooters"):
        writeFooterFunctions[parserName] = parserModules[parserName].writeFooters
    else:
        writeFooterFunctions[parserName] = False

def readFile(filename, filetype, parserArgs=False):
    """
    Reads the file with given filetype parser and parserArgs

    First, generates the reader objects from parserArgs. Then, splits into two possibilities:
    If line reader is present, iterates line by line. Starts with header reader, if present, until
    it throws ValueError. Then it continues to line reader. Each line should be converted to set of values
    of the same size for each line. False is returned if the line is to be skipped. ValueError is to
    be thrown if going to footer reader is required. Footer reader is the last stage. On the other hand,
    if line reader is not present, default to file reader. There, leave file opening to the parser, only parse
    filename and objects. Expect the entire dataset returned.
    """
    # Allow for default arguments
    if not parserArgs:
        parserArgs = readArgDefaults[filetype]
    # Initiate reading objects
    if initReadObjects[filetype]:
        readerObjects = initReadObjects[filetype](parserArgs.split())
    else:
        readerObjects = []
    # Check whether reading line by line is present, as it is preferred
    if readLineFunctions[filetype]:
        # read lines
        # requires text mode
        with open(filename, "r") as f:
            # Read headers, until False is produced
            if readHeaderFunctions[filetype]:
                for line in f:
                    try:
                        readHeaderFunctions[filetype](line, *readerObjects)
                    except ValueError:
                        # Value error signals that it is time to move to main content reading
                        break
            # Finished reading headers, read main content
            dataset = []
            for line in f:
                try:
                    datarow = readLineFunctions[filetype](line, *readerObjects)
                    if datarow:
                        # Either initiate dataset or append datarow
                        if len(datarow) > len(dataset):
                            for dataitem in datarow:
                                dataset.append([dataitem])
                        else:
                            for i in range(len(datarow)):
                                dataset[i].append(datarow[i])
                    else:
                        continue
                except ValueError:
                    # Time to read footers
                    break
            # Finished reading main content, read footers
            if readFooterFunctions[filetype]:
                for line in f:
                    readFooterFunctions[filetype](line, *readerObjects)
            # Done
        return dataset
    else:
        # Read whole file in one go
        return readFileFunctions[filetype](filename, *readerObjects)

def writeFile(filename, filetype, dataset, parserArgs=False):
    # Initialize parserArgs
    if not parserArgs:
        parserArgs = writeArgDefaults[filetype]
    # Start by initialising writer objects
    if initWriteObjects[filetype]:
        writerObjects = initWriteObjects[filetype](parserArgs.split())
    else:
        writerObjects = []
    # Differentiate between line-by-line writing and all-in-one writing
    if writeLineFunctions[filetype]:
        # Write line by line
        # If filename is false, write to stdout
        if not filename:
            f = sys.stdout
        else:
            f = open(filename, "w+")
        # Start by writing headers
        # Headers/footers can be written all-in-one always
        if writeHeaderFunctions[filetype]:
            headers = writeHeaderFunctions[filetype](dataset, *writerObjects)
            if len(headers) > 0:
                headers += "\n"
            f.write(headers)
        # Continue with lines
        for i in range(len(dataset[0])):
            datarow = []
            for j in range(len(dataset)):
                datarow.append(dataset[j][i])
            line = writeLineFunctions[filetype](datarow, *writerObjects)
            f.write(line+"\n")
        # Finish by writing footers
        if writeFooterFunctions[filetype]:
            footers = writeFooterFunctions[filetype](dataset, *writerObjects)
            if len(footers) > 0:
                footers += "\n"
            f.write(footers)
        if filename:
            f.close()
    else:
        # Write all-in-one
        writeFileFunctions[filetype](filename, dataset, *writerObjects)

def postProcess(datagroups, command, args):
   return postProcessCommands[command](datagroups, args)

def save(savepointGroup, context, datasets, defaultDatasetName=False):
    toSave = handleSavepoints(savepointGroup, context, defaultDatasetName)
    for savepointArgs in toSave:
        writeFile(savepointArgs["filename"], savepointArgs["parserName"], datasets[savepointArgs["datasetName"]], savepointArgs["parserArgs"])

def parseDatasetConfig(configFilename):
    cfg = configparser.ConfigParser()
    cfg.read(configFilename)
    datasets = {}
    for groupName in cfg.sections():
        if "dataset" in groupName:
            datasetName = groupName.split(".")[1]
            if "file" in cfg[groupName]:
                # Create datasets from file
                parserArgs = cfg[groupName].get("parser-args", False)
                datasets[datasetName] = readFile(cfg[groupName]["file"], cfg[groupName]["filetype"], parserArgs=parserArgs)
            elif "list" in cfg[groupName]:
                # Create dataset from list, defaultly convert to float
                # TODO : Should there be som interface to different convertors?
                # A more sophisticated and decoupled list types
                splitList = cfg[groupName]["list"].split("\n")
                if "listtype" in cfg[groupName]:
                    if cfg[groupName]["listtype"] == "string":
                        datasets[datasetName] = list(map(lambda s: s.split(), splitList))
                    else:
                        datasets[datasetName] = list(map(lambda s: list(map(float, s.split())), splitList))
                else:
                    datasets[datasetName] = list(map(lambda s: list(map(float, s.split())), splitList))
            if "savepoint" in cfg[groupName]:
                save(cfg[groupName].get("savepoint"), "load", datasets, datasetName)
            if "post-process" in cfg[groupName]:
                differentCommands = cfg[groupName].get("post-process").split("\n")
                for commandLine in differentCommands:
                    commandSplit = commandLine.split()
                    if len(commandSplit) > 1:
                        datasets[datasetName] = postProcess(datasets[datasetName], commandSplit[0], commandSplit[1:])
                    else:
                        datasets[datasetName] = postProcess(datasets[datasetName], commandSplit[0], [])
            if "savepoint" in cfg[groupName]:
                save(cfg[groupName].get("savepoint"), "post-process", datasets, datasetName)
    return datasets



import os
from .. import __user_conf_dir
import importlib

## Search for default post-processing commands
roots = []
modFilenames = []
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/transforms"))
roots.append(root)
modFilenames.append(filenames)
## Search for custom post_process commands
if os.path.isdir(os.path.expanduser(__user_conf_dir+"/transforms")):
    root, _, filenames = next(os.walk(os.path.expanduser(__user_conf_dir+"/transforms")))
    roots.append(root)
    modFilenames.append(filenames)
## Import all commands
transforms = {}
for i in range(len(roots)):
    for filename in modFilenames[i]:
        commandName = filename.split(".")[0]
        if commandName[0:2] == "__":
            # Skip __init__.py and similar commands
            continue
        spec = importlib.util.spec_from_file_location("compphysutils.graphics.transforms."+commandName, roots[i]+"/"+filename)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        transforms[commandName] = mod.command



import random
import configparser
from .readCrystalChar import readCrystalChar, parseCharConfig
from .VectorRepresentation import VectorRepresentation
from .VectorReal import VectorReal
from .. import __user_conf_dir
import os

def getUnitBasis(N):
    basis = []
    for i in range(N):
        basis.append(VectorRepresentation(*([0]*N)))
        basis[-1].components[i] = 1
    return basis

def checkPositionPresent(crystalRep, vecRep, basisRep):
    for otherRep in crystalRep:
        if vecRep.translateToBasis(basisRep) == otherRep.translateToBasis(basisRep):
            return True
    return False

def generateNextLevel(crystalRep, unitBasis, basisRep):
    """
    Grows the crystal by one layer.
    """
    # Iterate over positions of all atoms
    newLayer = []
    for vecRep in crystalRep:
        # Try adding each basis vector
        for unitRep in unitBasis:
            newVec = vecRep + unitRep
            if checkPositionPresent(crystalRep, newVec,basisRep) or checkPositionPresent(newLayer, newVec, basisRep):
                # Do not add this vector, skip it
                continue
            # Add the vector to the crystal
            newLayer.append(newVec)
    # The argument is augmented and returned
    return crystalRep + newLayer

def addAtoms(crystalRep, basisRep, planarBasis, addatoms, seed=False):
    if seed:
        random.seed(seed)
    for i in range(addatoms):
        # Choose atom at random
        atomIndex = random.randint(0,len(crystalRep)-1)
        newVec = crystalRep[atomIndex]
        # Create 1/-1 multiplier choices
        basisMultipliers = []
        for j in range(len(planarBasis)):
            basisMultipliers.append(random.choice([1,-1]))
        while checkPositionPresent(crystalRep, newVec, basisRep):
            basisIndex = random.randint(0, len(planarBasis)-1)
            newVec = newVec + (basisMultipliers[basisIndex] * planarBasis[basisIndex])
        crystalRep = crystalRep + [newVec]
    return crystalRep

def generateNLevels(N, unitBasis, basisRep):
    crystalRep = [VectorRepresentation(*([0]*len(unitBasis)))]
    for i in range(N-1):
        crystalRep = generateNextLevel(crystalRep, unitBasis, basisRep)
    return crystalRep

def getRealStructure(crystalRep, basisReal):
    crystal = []
    for vecRep in crystalRep:
        crystal.append(vecRep.translateToBasis(basisReal))
    return crystal

def getAtomData(crystal, unitLength, *atomicBasis):
    """
        Generates the atomic data based on the real crystal and the unit length
        Now updated to replicate the basis at each point of the crystal
        In order to do that, arguments are given as follows
        *args = [elemOneName, elemOneBasisVector, elemTwoName, elemTwoBasisVector, ...]
        if only single element is present, assume basis position is 0,0,0
    """
    data = []
    if len(atomicBasis) < 1:
        raise ValueError("Cannot assign elements to crystal points - no element name provided!")
    if len(atomicBasis) == 1:
        for pos in crystal:
            data.append(list((unitLength * pos).components) + [elemName])
    else:
        for pos in crystal:
            for i in range(0,len(atomicBasis),2):
                data.append(list((unitLength * (pos + atomicBasis[i+1])).components) + [atomicBasis[i]])
    return data

def generateCrystalRepresentation(basisRep, layers=1):
    unitBasis = getUnitBasis(len(basisRep))
    crystalRep = generateNLevels(layers, unitBasis, basisRep)
    # add-atoms
    #if not planarBasis:
    #    crystalRep = addAtoms(crystalRep, basisRep, unitBasis, addatoms, seed=seed)
    #else:
    #    crystalRep = addAtoms(crystalRep, basisRep, planarBasis, addatoms, seed=seed)
    return crystalRep, unitBasis

def generateCrystal(crystalCharFilename, layers=1, addatoms=0, seed=False):
    basisReal, basisRep, unitLength, planarBasis, atomicBasis = readCrystalChar(crystalCharFilename)
    crystalRep = generateCrystalRepresentation(basisRep, planarBasis, layers=layers, addatoms=addatoms, seed=seed)
    crystal = getRealStructure(crystalRep, basisReal)
    # Finally, convert to standard format - atomic data format and multiply by unitLength
    return getAtomData(crystal, unitLength, *atomicBasis)

def moveAtom(crystalRep, basisRep, index, vector):
    newVec = crystalRep[index] + vector
    if checkPositionPresent(crystalRep, newVec, basisRep):
        raise ValueError("Atom would be moved to an already occupied crystal site!")
    else:
        crystalRep[index] = newVec
    return crystalRep

def addAtomAtPosition(crystalRep, basisRep, vector):
    if checkPositionPresent(crystalRep, vector, basisRep):
        raise ValueError("Atom would be added to an already occupied crystal site!")
    else:
        crystalRep.append(vector)
    return crystalRep

def delAtomAtPosition(crystalRep, basisRep, vector):
    # Delete if possible, otherwise just raise warning, not exception
    deleted = False
    for i in range(len(crystalRep)):
        if vector == crystalRep[i]:
            del crystalRep[i]
            deleted = True
            break
    if not deleted:
        print("No atom found at position "+str(vector))
    return crystalRep

def delAtomAtIndex(crystalRep, index):
    del crystalRep[index]
    return crystalRep

def getDataString(atomicData):
    string = str(len(atomicData))+"\n\n"
    for i in range(len(atomicData)):
        string = string + atomicData[i][-1].upper()+" "+" ".join(map(str, atomicData[i][:3])) + "\n"
    return string

def printData(atomicData):
    print(getDataString(atomicData))

def saveData(atomicData, filename):
    file = open(filename, "w")
    file.write(getDataString(atomicData))
    file.close()

def postProcessVectorSum(unitBasis, planarBasis, instructions):
    vectorSum = unitBasis[0].cloneZeros()
    for i in range(0,len(instructions),2):
        if instructions[i] == "planar":
            vectorSum = vectorSum + planarBasis[int(instructions[i+1])]
        elif instructions[i] == "basis":
            vectorSum = vectorSum + unitBasis[int(instructions[i+1])]
        elif instructions[i] == "mbasis":
            vectorSum = vectorSum - unitBasis[int(instructions[i+1])]
        elif instructions[i] == "mplanar":
            vectorSum = vectorSum - planarBasis[int(instructions[i+1])]
    return vectorSum

def createClusterFromConfig(configFilename):
    cfg = configparser.ConfigParser()
    cfg.set("DEFAULT", "indexdump", "False")
    cfg.read(configFilename)
    if not (("lattice" in cfg["cluster"]) or (("base" in cfg.sections()) and ("vectors" in cfg.sections()))):
        raise ValueError("Missing configuration - need a lattice configuration file [cluster]->lattice or [base] and [vectors] sections in the config file.")
    else:
        if "lattice" in cfg["cluster"]:
            # Check whether the lattice exists in cwd or in config directory, cwd takes priority
            if os.path.isfile(cfg["cluster"]["lattice"]):
                basisReal, basisRep, unitLength, planarBasis, atomicBasis = readCrystalChar(cfg["cluster"]["lattice"])
            else:
                # Try config dir
                basisReal, basisRep, unitLength, planarBasis, atomicBasis = readCrystalChar(os.path.expanduser(__user_conf_dir+"/lattices/")+cfg["cluster"]["lattice"])
        else:
            basisReal, basisRep, unitLength, planarBasis, atomicBasis = parseCharConfig(cfg)
        crystalRep, unitBasis = generateCrystalRepresentation(basisRep, layers=int(cfg["cluster"]["layers"]), )
        # Remove the atoms if doing a sphere cut
        if "post-process" in cfg.sections():
            if "sphere-cut" in cfg["post-process"]:
                radius = float(cfg["post-process"].get("sphere-cut", 1.0))
                newCrystalRep = []
                for atom in crystalRep:
                    realPos = atom.translateToBasis(basisReal) * unitLength
                    if abs(realPos) <= radius:
                        newCrystalRep.append(atom)
                crystalRep = newCrystalRep
        # add-atoms
        if "addatoms" in cfg["cluster"]:
            if not planarBasis:
                crystalRep = addAtoms(crystalRep, basisRep, unitBasis, addatoms=int(cfg["cluster"]["addatoms"]), seed=cfg["cluster"].get("seed", False))
            else:
                crystalRep = addAtoms(crystalRep, basisRep, planarBasis, addatoms=int(cfg["cluster"]["addatoms"]), seed=cfg["cluster"].get("seed", False))
        # Start the post-process
        if "post-process" in cfg.sections():
            unitBasis = getUnitBasis(len(crystalRep[0].components))
            if "move" in cfg["post-process"]:
                for movementLine in cfg["post-process"]["move"].split("\n"):
                    movementInstructions = movementLine.split()
                    vectorSum = postProcessVectorSum(unitBasis, planarBasis, movementInstructions[1:])
                    # Final displacement vector complete - displace the atom
                    crystalRep = moveAtom(crystalRep, basisRep, int(movementInstructions[0]), vectorSum)
            # Movement instructions finished
            if "add" in cfg["post-process"]:
                for additionLine in cfg["post-process"]["add"].split("\n"):
                    additionInstructions = additionLine.split()
                    vectorSum = postProcessVectorSum(unitBasis, planarBasis, additionInstructions)
                    crystalRep = addAtomAtPosition(crystalRep, basisRep, vectorSum)
            if "delete" in cfg["post-process"]:
                for deletionLine in cfg["post-process"]["delete"].split("\n"):
                    deletionInstructions = deletionLine.split()
                    if deletionInstructions[0] == "index":
                        crystalRep = delAtomAtIndex(crystalRep, int(deletionInstructions[1]))
                    else:
                        vectorSum = postProcessVectorSum(unitBasis, planarBasis, deletionInstructions)
                        crystalRep = delAtomAtPosition(crystalRep, basisRep, vectorSum)
        structure = getRealStructure(crystalRep, basisReal)
        data = getAtomData(structure, unitLength, *atomicBasis)
        # If IndexDump is required, output atom data
        if "post-process" in cfg.sections():
            dumpIndexInfo = cfg.getboolean("post-process", "indexdump")
            if dumpIndexInfo:
                for i in range(len(crystalRep)):
                    print(i, crystalRep[i], data[i])
                if(planarBasis):
                    print("Planar basis")
                    for i in range(len(planarBasis)):
                        print(i, planarBasis[i])
        if "post-process" in  cfg.sections():
            if "random-noise" in cfg["post-process"]:
                # Generate random Gaussian noise with given amplitude for each atom in the system
                # This is applied at the very end in order to allow manipulation of crystal before
                argSplit = cfg["post-process"].get("random-noise").split()
                radius = float(argSplit[0])
                seed = int(argSplit[1])
                random.seed(seed)
                for i in range(len(data)):
                    for j in range(3):
                        data[i][j] += random.gauss(0,radius)
        saveData(data, cfg["cluster"]["name"])



from scipy.optimize import curve_fit
from .. import __user_conf_dir
import math
import importlib
import os

roots = []
modFilenames = []
root, _, filenames = next(os.walk(os.path.dirname(__file__)+"/fit_types"))
roots.append(root)
modFilenames.append(filenames)
## Search for custom post_process commands
if os.path.isdir(os.path.expanduser(__user_conf_dir+"/fit_types")):
    root, _, filenames = next(os.walk(os.path.expanduser(__user_conf_dir+"/fit_types")))
    roots.append(root)
    modFilenames.append(filenames)
## Import all commands
fitFunctions = {}
guessFunctions = {}
paramNames = {}
for i in range(len(roots)):
    for filename in modFilenames[i]:
        fitFuncName = filename.split(".")[0]
        if fitFuncName[0:2] == "__":
            # Skip __init__.py and similar commands
            continue
        spec = importlib.util.spec_from_file_location("compphysutils.graphics.fit_types."+fitFuncName, roots[i]+"/"+filename)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        fitFunctions[fitFuncName] = mod.fit
        paramNames[fitFuncName] = mod.paramNames
        if hasattr(mod, "guess"):
            guessFunctions[fitFuncName] = mod.guess
        else:
            guessFunctions[fitFuncName] = False

def roundSignificantFigures(number, sigFigs, matchOrder=False):
    if not matchOrder:
        # Default behaviour - match to given number of significant figures
        try:
            result = round(number, sigFigs - 1 - math.floor(math.log10(abs(number))))
        except OverflowError:
            # Number is infinity, just output infinity
            result = "Infinity"
    else:
        # Match to the order of the sigFigs argument
        # Problem is that the variance may be very high
        try:
            result = round(number, -math.floor(math.log10(abs(sigFigs))))
        except OverflowError:
            # Infinity - return number rounded to one significant figure
            try:
                result = round(number, -math.floor(math.log10(abs(number))))
            except OverflowError:
                # If even this overflows, number is infinity, can just output infinity
                result = "Infinity"
    return result

def plotFit(dataset, fitFunctionName, axisObj, **fitParams):
    # Behaviour changes depending on the number of columns
    # TODO : Do other possibilities (i.e. xerr and yerr and no error)
    # Find the indices for the required coordinates
    ixMin = 0
    if fitParams["xMin"]:
        while fitParams["xMin"] > dataset[0][ixMin]:
            ixMin += 1
    ixMax = len(dataset[0])-1
    if fitParams["xMax"]:
        while fitParams["xMax"] < dataset[0][ixMax]:
            ixMax -= 1
    # Guess the initial params for faster fitting (or succesfull fitting at all)
    guesses = None
    if guessFunctions[fitFunctionName]:
        guesses = guessFunctions[fitFunctionName](dataset[0][ixMin:ixMax+1], dataset[1][ixMin:ixMax+1])
    if fitParams["dirtyRun"]:
        popt = guesses
        perr = guesses
    else:
        if len(dataset) == 3:
            popt, pcov = curve_fit(fitFunctions[fitFunctionName], dataset[0][ixMin:ixMax+1], dataset[1][ixMin:ixMax+1], sigma=dataset[2][ixMin:ixMax+1], p0=guesses)
        else:
            popt, pcov = curve_fit(fitFunctions[fitFunctionName], dataset[0][ixMin:ixMax+1], dataset[1][ixMin:ixMax+1], p0=guesses)
        perr = []
        for i in range(len(pcov)):
            perr.append(pcov[i][i] ** 0.5)
    xMin = dataset[0][ixMin]
    xMax = dataset[0][ixMax]
    dx = (xMax - xMin) / (fitParams["fitPoints"] - 1)
    xs = []
    ys = []
    for i in range(fitParams["fitPoints"]):
        xs.append(xMin + dx*i)
        ys.append(fitFunctions[fitFunctionName](xMin + dx*i, *popt))
    if fitParams["fitLabel"]:
        axisObj.plot(xs,ys,label=fitParams["fitLabel"],color=next(fitParams["fitColorCycle"]))
    else:
        axisObj.plot(xs,ys,color=next(fitParams["fitColorCycle"]))
    # Construct the param string
    if fitParams["showParams"]:
        pstring = ""
        for i in range(len(popt)):
            result = roundSignificantFigures(popt[i], perr[i], matchOrder=True)
            error = roundSignificantFigures(perr[i], 1)
            if fitParams["showError"]:
                pstring += paramNames[fitFunctionName][i]+" : "+str(result)+r"$\pm$"+str(error)+"\n"
            else:
                pstring += paramNames[fitFunctionName][i]+" : "+str(result)+"\n"
        if fitParams["paramsPlacement"]:
            # Text anchor is the bottom left corner by default
            if fitParams["paramsPlacement"] == "tl":
                axisObj.text(0.1, 0.9-0.05*(len(popt)-1)-0.05*(fitParams["paramsOffset"]), pstring, transform=axisObj.transAxes)
        else:
            # Default to top left
            axisObj.text(0.1, 0.9-0.05*(len(popt)-1)-0.05*(fitParams["paramsOffset"]), pstring, transform=axisObj.transAxes)
    return popt



shortContexts = ["load", "post-process"]

def parseArgs(savepointArgString, context, defaultDatasetName=False):
    result = {}
    savepointArgs = savepointArgString.split()
    result["context"] = savepointArgs[0]
    parserArgsStartIndex = 3
    # If the context is short, expect no dataset name - it needs to be provided by program
    if result["context"] in shortContexts:
        if not defaultDatasetName:
            raise ValueError("Invoking savepoint at short context but not providing default dataset name!")
        else:
            # Dataset name defined, proceed with parsing
            result["datasetName"] = defaultDatasetName
            result["parserName"] = savepointArgs[1]
            # Change parser start index
            parserArgsStartIndex = 2
    else:
        # Lond context
        result["datasetName"] = savepointArgs[1]
        result["parserName"] = savepointArgs[2]
    # Check for index before filename
    try:
        filenameSeparatorIndex = savepointArgs.index("--")
    except ValueError:
        filenameSeparatorIndex = -1
    if filenameSeparatorIndex == -1:
        # No filename, go to default filename
        result["filename"] = result["context"] + ".out"
        # If there is a single extra arg after parserName, assume it is also fileName
        if len(savepointArgs) == parserArgsStartIndex+1:
            result["filename"] = savepointArgs[parserArgsStartIndex]
            filenameSeparatorIndex = parserArgsStartIndex
    else:
        result["filename"] = savepointArgs[filenameSeparatorIndex+1]
    # Finally, determine parser args
    if filenameSeparatorIndex != -1:
        parserArgs = savepointArgs[parserArgsStartIndex:filenameSeparatorIndex]
    else:
        parserArgs = savepointArgs[parserArgsStartIndex:]
    result["parserArgs"] = parserArgs
    return result

def handleSavepoints(savepointGroup, context, defaultDatasetName=False):
    results = []
    for savepointArgs in savepointGroup.split("\n"):
        # Get parser arguments
        partialResult = parseArgs(savepointArgs, context, defaultDatasetName=defaultDatasetName)
        if partialResult["context"] == context:
            results.append(partialResult)
    return results




import argparse
import numpy

imageAP = argparse.ArgumentParser("Plots the image loaded as dataset to a given axis coordinate dimension.")
imageAP.add_argument("imageDataset", help="Name of the dataset containing the RGB(A) data. They should be in an unrolled loop.")
imageAP.add_argument("imageRect", nargs=4, type=float, help="Rectangle for the image to be drawn into, in axes coordinates. (x,y,width,height).")
imageAP.add_argument("--shape", type=lambda x: list(map(int, x.split(","))), help="Shape of the image, in format width,height.")

def command(axes, datasets, argString):
    args = imageAP.parse_args(argString)
    numberOfChannels = len(datasets[args.imageDataset])
    # Shape the image
    shapedImage = numpy.array(datasets[args.imageDataset]).reshape([numberOfChannels] + args.shape)
    # Transform to correct shape for plotting
    shapedImage = numpy.moveaxis(shapedImage, 0, -1)
    # Plot the image, with transformation of the coordinates
    axes.imshow(shapedImage, transform=axes.transAxes, aspect="auto", origin="lower", extent=(args.imageRect[0],args.imageRect[0]+args.imageRect[2],args.imageRect[1],args.imageRect[1]+args.imageRect[3]))
    return axes, datasets
